(*  Converted from Haskell to CaML (Sun Aug 16 13:05:14 EST 1998) 
\chapter{QProcessor}

This module acts a bit like a quantum processor does.  There are two
major functions
\begin{description}
\item[evaluate] simulates the result of applying a particular gate
operation
\item[observe] simulates the result of observing the system
\end{description}

\section{Preliminaries}

*)

module QProcessor where

open GregComplex
open Complex
open Type
open LUtable

(*

\section{Types}

*)
type ProcessorState = LUtable Integer Type
type Processor = [(ProcessorState,GregComplex)]
type GateOperation = [Type] -> [([Type],GregComplex)]
(*

\section{Initials}

Some values are going to be needed in other parts of the program.  
For example,  I'm going to need get my first register pool from
somewhere\footnote{Actually,  I should make it return an opaque enumerate
type.  I can't remember why I didn't just now.}.  Similarly,  I'm
going to need to get a fresh processor from somewhere.

*)
fresh_Register_pool :: [Integer]
start_processor :: Processor


fresh_Register_pool = [1000..]
start_processor = [(new_LUtable,1)]
(*

\section{Public Functions}

\subsection{Evaluate}

Comments pending.  I've modified shrinkify  since modified doesn't
do a sort any more.

*)
    
evaluate :: Processor -> GateOperation -> [Integer] -> [Integer] -> Processor
evaluate proc oper inputs outputs = proc_alt
 where proc_alt = shrinkify modified
       modified = [(state,local_amplitude * group_amplitude) |
                      (original_state,group_amplitude) <- proc , 
                      (state,local_amplitude)
                                  <- doGateOperation original_state
                                                   oper inputs outputs
                   ]
       shrinkify [] = []
       shrinkify ((t,c):ts) = (t,total): shrinkify unlike_minded
         where like_minded = filter (\(x,y) -> x == t) ts
               unlike_minded = filter (\(x,y) -> x /= t) ts
               total = (sum . snd . unzip) ((t,c):like_minded)

(*

\subsection{Observe}

\texttt{observe} calculates the mod squares of the amplitudes of its input
states. (i.e. it collapses the wave function of the quantum processor)
If the sum of these mod squares is not {\em very} close to unity then
something more drastic than a loss of numeric precision has occurred.
Probably, this means there is a bug (somewhere) in this program.

\texttt{observe} expects to be given a real number between 0 and 1 (0
inclusive). This is its random-number-from-God --- this is the spooky
quantum indeterminancy thing which lots of people believe underlies 
the possibility of free will in human thought.  See Penrose~\cite{Penrose}
talks about that sort of thing.

\texttt{observe} uses that random number to choose one of the
ProcessorStates.  It was asked to observe on a particular wire,  so 
it finds out what value the wire had in that ProcessorState,  and then
returns a Processor containing just those ProcessorStates that had the
wire at that value.  It also then scales their amplitudes so that
the mod square sum is one again.

*)        
observe:: Processor -> Integer -> Double -> (Type,Processor)
observe hilbert which amount
 | not metrified = error ("observe --> metric = " 
                       ++ show metric ++ "\n"++ printProcessor hilbert)
 | not sensible_amount = error ("observe --> amount = "
                       ++ show amount)
 | otherwise = (observation,factored)
   where collapsed = [(a, (Complex.magnitude b)^2) | (a,b) <- hilbert]
         metric = (sum . snd . unzip) (collapsed)
         
         metrified = (abs (metric - 1.0)) < 0.000001
         sensible_amount = (amount >= 0.0) && (amount < 1.0)
         
         selection _ [] = error "bizarre selection"
         selection v ((s,c):ss)
          | w <= 0.0 = s
          | otherwise = selection w ss
            where w = v - c
            
         choice = selection amount collapsed
         observation = use_ret (valueOf choice which)
         factored = [(s,c / toGregComplex factor)
                     | (s,c) <- hilbert ,
                      use_ret (valueOf s which) == observation]
         factor = sqrt (sum [c | (b,c) <- collapsed ,
                                 use_ret (valueOf b which) == observation] )
                      
                      

                
            
(*

\section{Private Functions}

All these things beginning with ``print'' could (and should) be turned
into Show instances.  Concurrent Clean got a bit stroppy about certain
kinds of  {\tt toString} operations,  so I had to work around it.
Now that I'm using Haskell,  this isn't a problem.

*)
printProcessor :: Processor -> String
printProcessor [] = "\n"
printProcessor (x:xs) = printProcessorState x ++ "  " ++ 
                          printProcessor xs

printProcessorState :: (ProcessorState,GregComplex) -> String
printProcessorState (state,amplitude) =
 show amplitude ++ " | " ++ printEntries (toList state)

printEntries :: [(Integer,Type)] -> String
printEntries [] = ">"
printEntries (x:xs) = printEntry x ++ "," ++ printEntries xs

printEntry :: (Integer,Type) -> String
printEntry (x,y) = show x ++ "->" ++ show y
(*

{\tt doGateOperation} gets the values of the inputs in a given processor 
state,  performs the gate operation on those inputs (which returns a list
of (output,amplitude pairs)) and places those back into the output
registers.  Oh,  and it also destroys the input registers so that we
can't tell by what path we reached the current state.  

Note that what it returns is a {\em Processor}. One day I'm going
to get really swish and say that some exciting kind of Monadic
operation is going on here.

*)
doGateOperation :: ProcessorState -> GateOperation -> [Integer] -> [Integer] -> Processor
doGateOperation old_state oper inputs outputs = ret
 where f = [ use_ret (valueOf old_state r) | r <- inputs ]
       f_alt = oper f
       simpler_state = use_ret (removeMany old_state inputs)
       ret = [(insertList simpler_state (zip outputs outs),c) | (outs,c) <- f_alt]
            
(*
*)
