(*  Converted from Haskell to CaML (Sun Aug 16 13:05:14 EST 1998) 
 qprocessor.ml

\chapter{QProcessor}

This module acts a bit like a quantum processor does.  There are two
major functions
\begin{description}
\item[evaluate] simulates the result of applying a particular gate
operation
\item[observe] simulates the result of observing the system
\end{description}
\section{Types}

*)
open Complex

(* processor_state *)
(* I'm not quite happy thinking about the performance of this.... *)

type processor_state = (int * Qgol_type.t) list;;

let procstate_sort procstate=
  let processor_state_sort_function (i1,t1) (i2,t2) = i1 <= i2  in
   Sort.list processor_state_sort_function procstate;;

let procstate_equal x y =
  procstate_sort x = procstate_sort y;;

let procstate_show procstate = 
  "|" ^ 
  (Util.string_join "; " 
    (List.map (function (x,y)-> ((string_of_int x) ^ "=" ^ (Qgol_type.show y)))
      procstate
    )
  ) ^ ">";;

(* processor *)

type processor = (processor_state * complex) list;;

let normalised proc =
 let (_,amplitudes) = List.split proc in
 let probabilities = List.map mod_square amplitudes in
 let probability_sum = List.fold_left (+.) 0.0 probabilities
in 
 abs_float (probability_sum -. 1.0) < 0.0000001
;;

let show_processor proc = 
  (Util.string_join " +  " 
    (List.map (function (proc_state,amplitude) -> 
         (show amplitude) ^ (procstate_show proc_state) ) proc) )
;;
(* gate_operation *)

type gate_operation = Qgol_type.t list -> ((Qgol_type.t list * complex) list);;


(*           
{\tt doGateOperation} gets the values of the inputs in a given processor 
state,  performs the gate operation on those inputs (which returns a list
of (output,amplitude pairs)) and places those back into the output
registers.  Oh,  and it also destroys the input registers so that we
can't tell by what path we reached the current state.  

Note that what it returns is a {\em Processor}. One day I'm going
to get really swish and say that some exciting kind of Monadic
operation is going on here.

*)

let doGateOperation group_amplitude old_state oper inputs outputs =
  let f = List.map (function r -> List.assoc r old_state) inputs
  and simpler_state = Util.filter (function (x,y) -> not (List.mem x inputs)) old_state in

    List.map (function (new_values,amplitude) -> 
                         (simpler_state @ (List.combine outputs new_values),
                          group_amplitude &*& amplitude)
                                                         ) (oper f)


(*

\section{Initials}

Some values are going to be needed in other parts of the program.  
For example,  I'm going to need get my first register pool from
somewhere\footnote{Actually,  I should make it return an opaque enumerate
type.  I can't remember why I didn't just now.}.  Similarly,  I'm
going to need to get a fresh processor from somewhere.

*)

(*

I want to do something like this:

let fresh_Register_pool = [1000..]

but I can't work out how to get lazy structures properly.
I should be able to calculate how many registers I will need,
actually,  but I haven't had time to do it.  So for now...
*)
let fresh_Register_pool = Util.list_range 1000 2000



let start_processor = [([],{re=1.0;im=0.0})]
(*

\section{Public Functions}

\subsection{Evaluate}

*)
    

(* simplify_processor looks through the set of states of a processor
and looks for a processor states that are the same.  If it finds any,
it adds up their amplitudes and stores the result *)

let simplify_processor proc =
 let sorted_processor_states = 
       List.map (function (proc_state,amplitude)  
             -> (procstate_sort proc_state,amplitude)  ) proc  
 in let rec amplitude_sum pr = 
   match pr with 
     [] -> {re=0.0;im=0.0}
   | (_,amplitude)::others -> amplitude &+& (amplitude_sum others)

 and simplify sts =
   match sts with
     [] -> []
   | (proc_state,amplitude)::other_waves -> 
       let similar_state_identifier (p,a) = procstate_equal proc_state p in
       let (like,unlike) = Util.partition similar_state_identifier other_waves
       in (proc_state,amplitude_sum ((proc_state,amplitude)::like))::(simplify unlike)
 in
  simplify sorted_processor_states
;;


(* [evaluate proc oper inputs outputs] applies an ordinary gate_operation
   [oper] to a processor [proc].  [oper] will (obviously) only operate
   on some of the wires [inputs] present in [proc],   and the outputs
   should go to the wires specified by [outputs] *)


let evaluate proc oper inputs outputs = 
 let 
  new_processor_states =
   List.map (function (original_state,group_amplitude)
               -> doGateOperation group_amplitude original_state oper inputs outputs)
              proc
 in  simplify_processor (List.flatten new_processor_states);;

(*

\subsection{Observe}

\texttt{observe} calculates the mod squares of the amplitudes of its input
states. (i.e. it collapses the wave function of the quantum processor)
If the sum of these mod squares is not {\em very} close to unity then
something more drastic than a loss of numeric precision has occurred.
Probably, this means there is a bug (somewhere) in this program.

\texttt{observe} creates a real number between 0 and 1 (0
inclusive). This is its random-number-from-God --- this is the spooky
quantum indeterminancy thing which lots of people believe underlies 
the possibility of free will in human thought.  See Penrose~\cite{Penrose}
talks about that sort of thing.

\texttt{observe} uses that random number to choose one of the
ProcessorStates.  It was asked to observe on a particular wire,  so 
it finds out what value the wire had in that ProcessorState,  and then
returns a Processor containing just those ProcessorStates that had the
wire at that value.  It also then scales their amplitudes so that
the mod square sum is one again.

*)        



let observe hilbert which =
 Multiple_results.fork_off () ;
 let amount = Random.float 1.0 in
 if (not (normalised hilbert)) then 
  (prerr_endline ("observe was called with a hilbert space of " ^ (show_processor hilbert) ^ ". Exitting.") ; exit 1) else


 let rec selection_func v states =
      match (v,states) with
        (_,[]) -> (prerr_endline ("bizarre selection") ; exit 1)
      |	(v,(s,p)::ss) ->
          if (v <= p) then s else selection_func (v -. p) ss
 in let collapsed_space = 
      List.map (function (proc,amplitude) -> (proc, mod_square amplitude))
              hilbert
 in
 let choice = selection_func amount collapsed_space in
 let observation = List.assoc which choice in
 let such_observations (s,q) = List.assoc which s = observation in
 let surviving_hilbert = Util.filter such_observations hilbert and
     (_,survive_ps) = List.split (Util.filter such_observations collapsed_space) in
 let factor = {re = 1.0 /. (sqrt ( List.fold_right (+.) survive_ps 0.0 ));
		im = 0.0 } in
 let factored = List.map (function (s,a) -> (s, a &*& factor)) surviving_hilbert
     in (observation,factored)
;;

                
	
